#~/usr/bin/python
import AGG,math


RESERVE = 1

def toRevenueFormat(setting):
    n,m,k,beta,V,P = setting
    V = [ ( [0.0]*n ) ]*m
    return (n,m,k,beta,V,P)


def __unweight(setting):
    n,m,k,beta,V,P = setting
    beta = [1.0]*n
    return (n,m,k,beta,V,P)

GSP = "GSP"
GFP = "GFP"
WEIGHTED = "WEIGHTED"
UNWEIGHTED = "UNWEIGHTED"
PPC = "PPC"
PPI = "PPI"
REVENUE_TEST = "REVENUE"
NORMALIZE = "NORMALIZE"

AUCTION_OPTIONS = [GSP ,GFP ,WEIGHTED ,UNWEIGHTED ,PPC,PPI ,REVENUE_TEST]

def makeAuction(setting,auction):
    assert GSP in auction or GFP in auction
    assert WEIGHTED in auction or UNWEIGHTED in auction
    assert PPC in auction or PPI in auction
    
    if UNWEIGHTED in auction:
        setting = __unweight(setting)
        
    n,m,k,beta,V,P = setting

    # Players
    N = range(n)

    A,S,E = makeActionNodes(N,setting,PPC in auction)
    	
    Fn = makeFunctionNodes(E,GSP in auction)

    v = makeArcs(A,E,GSP in auction)

    f = makeFunctionTypes(Fn)
    if REVENUE_TEST in auction:
        setting = toRevenueFormat(setting)
    if GFP in auction:
        u = makeUtilityFunctionsForGFP(setting,A,PPC in auction)
    else:
        u = makeUtilityFunctionsForGSP(setting,A,Fn,PPC in auction)

    return AGG.AGG(N,A,S,Fn,v,f,u)

def prune(N,S,A,setting,ppc):
	n,m,k,beta,V,P = setting
	# Prune bids above max value 
	for i in N:
		if ppc:
			maxV = math.ceil(max([(V[j][i]) for j in range(m)]))
		else:
			maxV = math.floor(max([(P[j][i]*V[j][i]) for j in range(m)]))
		remove = []
		for a in S[i]:
			if a[3] > maxV:
				remove.append(a)
		for a in remove:
			S[i].remove(a)

def value(setting,bidder,perClick):
    n,m,k,beta,V,P = setting
    if perClick:
        return max([(V[j][bidder]) for j in range(m)])
    else:
        return max([(P[j][bidder]*V[j][bidder]) for j in range(m)])
    
def makeActionNodes(N,setting,perClick):
    n,m,k,beta,V,P = setting
    A = []
    S = {}
    for i in N:
        v = int(math.floor(value(setting,i,perClick)))
        S[i] = [(BD,beta[i]*b,i,b) for b in range(v+1)]
        A = A + S[i]
    A.sort(lambda a1,a2: cmp(a1[BID],a2[BID]))
    newA = [A[0]]
    for i in range(1,len(A)):
        a1 = A[i-1]
        a2 = A[i]
        if a1[2]!=a2[2]:
            newA.append(a2)
        else:
            S[a2[2]].remove(a2)
    A = newA
    # Effective bids (very useful elsewhere)
    E = []
    for a in A:
        if a[BID] not in E: E.append(a[BID])
        
    return A,S,E

def makeFunctionNodes(E,priceNodes):
    Fn = [(EQ,e) for e in E]+[(GR,e) for e in E]
    if priceNodes: Fn+=[(PR,e) for e in E]
    return Fn

def makeArcs(A,E,priceNodes):
    # Every action node connects to/from an effective bid counter
    v = [( (EQ,a[BID]),a ) for a in A] 
    v += [( a,(EQ,a[BID]) ) for a in A]

    # Effective bid counter feeds into greater than counter
    v += [( (EQ,e),(GR,e) ) for e in E]

    # Greater than counter feeds into next highest greater than counter
    v += [( (GR,E[i+1]),(GR,E[i]) ) for i in range(len(E)-1)]

    # Greater than counter feeds into action node
    v += [( (GR,a[BID]),a ) for a in A]

    # Price arcs
    if priceNodes:
        # Price for an action feeds into that action
        v += [( (PR,a[BID]), a) for a in A]

        # For every price node, all bids below that feed into it
        for e in E:
            for a in A:
                if a[BID] >= e: break
                v.append(  (a,(PR,e))  )
    return v

def makeFunctionTypes(Fn):
    types = {}
    for fn in Fn:
        if fn[0] == PR:
            types[fn] = MAX
        else:
            types[fn] = SUM
    return types

def payoff(setting,i,pos,price,perClick):
    n,m,k,beta,V,P = setting    
    if pos >= m: return 0.0
    payoff = P[pos][i]*V[pos][i] # Expected value of clicks
    if perClick:
        payoff -= P[pos][i]*price
    else:
        payoff -= price
    return payoff

def makeUtilityFunctionsForGFP(setting,A,perClick):
    n,m,k,beta,V,P = setting    
    u = {}
    for aIndex in range(len(A)):
        a = A[aIndex]
        x,bid,i,b = a
        del x
        u[a] = {}
        ties = len(filter(lambda action: action[BID] == bid,A))
        for e in range(1,ties+1):
            for g in range(e,n+1):
                u[a][(e,g)] = 0.0
                if b == 0: continue
                ''' This price should never affect anything, as it implies
                that the lowest bidder has won something '''
                for pos in range(g-e,g):
                    u[a][(e,g)] += payoff(setting,i,pos,b,perClick)
                u[a][(e,g)]/=float(e)
    return u

def makeUtilityFunctionsForGSP(setting,A,Fn,perClick):
    n,m,k,beta,V,P = setting
    u = {}
    nodes = len(A)+len(Fn)
    for aIndex in range(len(A)):
        a = A[aIndex]
        x,bid,i,b = a
        del x
        u[a] = {}
        ties = len(filter(lambda action: action[BID] == bid,A))
        for e in range(1,ties+1):
            for g in range(e,n+1):
                for p in range(aIndex+1)+[nodes]:
                    u[a][(e,g,p)] = 0.0
                    if b == 0: continue
                    ''' This price should never affect anything, as it implies
                    that the lowest bidder has won something '''
                    if p == nodes: price = -10000
                    else: 
                        price = A[p][BID]/beta[i]
                    price = max(math.ceil(price),RESERVE)
                    #if p == 0: price = -10000
                    for pos in range(g-e,g-1):
                        u[a][(e,g,p)] += payoff(setting,i,pos,b,perClick)
                    u[a][(e,g,p)] += payoff(setting,i,g-1,price,perClick)
                    u[a][(e,g,p)]/=float(e)
    return u


EQ = 'equal to'
GR = 'greater or equal'
PR = 'price for'
BD = 'bid of'

MAX = 2
SUM = 0

BID = 1


def entropy(probabilityVector):
    probabilityVector = filter(lambda x: x!=0.0,probabilityVector)
    return -sum([(p*math.log(p,2)) for p in probabilityVector])

def entropyOfProfile(S):
    return sum([(entropy(S[i].values())) for i in S.keys()])

              
def testSimpleAuction(n,m,k,seed):
    import Setting
    setting = Setting.generateSimple(n,m,k,seed)

    setting = Setting.normalizePerClick(setting)
    auction = [GSP,WEIGHTED,PPC]
    agg = makeAuction(setting,auction)
    #print setting
    falseAGG = makeAuction(setting,auction+[REVENUE_TEST])
    falseAGG.baseName = ".false"
    falseAGG.saveToTextFile(".false.agg")
    nes= agg.solve()

    for ne in nes:
        print "Nash equilibrium of game: ",ne
        for i in agg.N:
            print i,":",
            for a ,p in ne[i].items():
                print (a[3],p),
                print
        print "Payoffs:",agg.payoffs(ne)
        print "Entropy:",entropyOfProfile(ne)
        surplus = agg.socialWelfare(ne)
        revenue = -falseAGG.socialWelfare(ne)
        print "Social Welfare",revenue+surplus,"Revenue",revenue
        socWelf = Setting.socialWelfare(setting)
        revenue = Setting.VCG_Revenue(setting)
    print "VCG Social Welfare",socWelf,"Revenue",revenue
    for g in range(setting[1]):
        print 'Values',[(setting[4][g][i]*setting[5][g][i]) for i in range(setting[0])]     


if __name__=="__main__":
    import Setting
    from sys import argv
    try:
        settingFilename = argv[1]
        aggFilename = argv[2]
        options = argv[3:]
        for option in options:
            assert option in AUCTION_OPTIONS
        setting = Setting.read(settingFilename)
        agg = makeAuction(setting,options)
        agg.saveToTextFile(aggFilename)
    except:
        print "usage: %s <setting filename> <agg filename> [auction options]"%(argv[0])
        print "Where auction options are:",AUCTION_OPTIONS
        
