import random, string,re,math,os
import Setting
from sys import argv
from AGG import PayoffComputer

random.seed(1)

def tweakNE(pc,neStr,objectiveFN):
    actions = [(i,j,d) for i in range(len(pc.aSizes)) for j in range(pc.aSizes[i]) for d in [-1,+1]]
    random.shuffle(actions)
    print "Stepping",
    for i,a1,direction in actions:
        print ".",
        ne = pc.parse(neStr)
        a2 = a1+direction
        if ne[i][a1] == 0.0: continue
        if a2 < 0 or a2 >= len(ne[i]): continue
        ne[i][a2] += ne[i][a1]
        ne[i][a1] = 0.0
        out = ""
        for i in range(len(ne)):
            for a in range(len(ne[i])):
                out += str(ne[i][a])+" "
        if objectiveFN(out) > objectiveFN(neStr):
            if pc.isNE(out):
                print
                return out
    print
    return None

def localSearch(pc,nes,objectiveFN):
    bestScore,bestNE = max([(objectiveFN(ne),ne) for ne in nes])
    for start in nes:
        for i in range(10): # Two restarts, does the second help?
            #print "\t",string.strip(start),i,objectiveFN(start)
            ne = start
            print "Start"
            while True:
                print objectiveFN(ne)
                new = tweakNE(pc,ne,objectiveFN)
                if new == None:
                    break
                ne = new
                #print ne,objectiveFN(ne)
            if objectiveFN(ne) > bestScore:
                bestScore,bestNE = objectiveFN(ne),ne
    return bestNE

def findExtremeNEs(pc1,pc2,nes):
   objectives = [lambda x:testNE(pc1,pc2,x)[0],lambda x:-testNE(pc1,pc2,x)[0],
   lambda x:testNE(pc1,pc2,x)[1],    lambda x:-testNE(pc1,pc2,x)[1] ]
   objectiveNames = ["max revenue","min revenue","max welfare","min welfare"]
   newNEs = []
   for objective,name in zip(objectives,objectiveNames):
       print "Search for",name
       newNEs.append(localSearch(pc1,nes,objective))
   return newNEs

''' FIXME: This looks like it's used elsewhere '''

def makePCs(prefix):
	os.system("make "+prefix+".agg")
	os.system("make "+prefix+"_Revenue.agg")
        pc1 = PayoffComputer(prefix+".agg")
        pc2 = PayoffComputer(prefix+"_Revenue.agg")
	return pc1,pc2

def entropy(probs):
	probs = filter(lambda x:x>0.0,probs)
	return -sum([(x*math.log(x)) for x in probs])

def support(probs): return len(filter(lambda x:x>0.0,probs))

def testNE(pc1,pc2,ne):
    #print ne
    payoffs = pc1.test(ne)
    payments = pc2.test(ne)
    revenue = -sum(map(float,string.split(payments)))
    welfare = sum(map(float,string.split(payoffs)))+revenue
    entropies = map(entropy,pc1.parse(ne))
    supports = map(support,pc1.parse(ne))
    return revenue,welfare,sum(entropies),max(entropies),sum(supports),max(supports)

def valid_strategy_profile(s):
    s = string.strip(s)
    parts = re.split("[ ,]+",s)
    for part in parts:
        try: float(part)
        except: return False
    return True

def load_GNM_NEs(filename):
    lines = open(filename,'r').readlines()
    runLineNumbers = filter(lambda i:lines[i]=="run #0\n", range(len(lines)-1))
    solutions = [(lines[i+1]) for i in runLineNumbers]
    solutions = filter(valid_strategy_profile,solutions)
    return solutions
    
if __name__=="__main__":
    try:
        if argv[2] in ["VCG","VCG_Discrete"]:
            assert len(argv)==3
        else:
            assert len(argv)==6
    except:
        print "usage: %s <setting filename> <agg filename> <revenue agg filename> <simpdiv equilibrium filename> <gnm equilibrium filename>"%(argv[0])
        print "or: %s <setting filename> VCG"%(argv[0])
        print "or: %s <setting filename> VCG_Discrete"%(argv[0])
        exit(1)

    if True:
        settingFilename = argv[1]
        setting = Setting.read(settingFilename)
        if argv[2] == "VCG":
            revenues = [Setting.VCG_Revenue(setting)]
            welfares = [Setting.socialWelfare(setting)]
        elif argv[2] == "VCG_Discrete":
            setting2 = Setting.DiscretizeForVCG(setting)
            revenues=[Setting.VCG_Revenue(setting2)]
            oa = Setting.optimalAssignment(setting2)
            welfares=[Setting.socialWelfare(setting,oa)]
        else:
            aggFilename = argv[2]
            aggRevenueFilename = argv[3]
            simpdivFilename = argv[4]
            gnmFilename = argv[5]
            pc1 = PayoffComputer(aggFilename)
            pc2 = PayoffComputer(aggRevenueFilename)
            nes = open(simpdivFilename,'r').readlines()
            nes = nes+load_GNM_NEs(gnmFilename)
            distinct = list(set(nes))
            print "#equilibria:",len(nes),"#distinct:",len(distinct)
            nes = distinct
                            
            if nes:
                nes2 = findExtremeNEs(pc1,pc2,nes)
                nes+=nes2
                outcomes = [(testNE(pc1,pc2,ne)) for ne in nes]
                revenues= [(outcome[0]) for outcome in outcomes]
                welfares= [(outcome[1]) for outcome in outcomes]
            else:
                revenues = ["N/A"]
                welfares = ["N/A"]
        print "Min_Revenue:",min(revenues), "Min_Social_Welfare:",min(welfares),"Max_Revenue:",max(revenues), "Max_Social_Welfare:",max(welfares)
