import string, math, random
'''

	A setting in an AGG:
n agents
m slots
k vector of bids
V matrix of values per click |n x m|
P matrix of click probabilities |n x m|   # FIXME: Change this to C for Click-through
B vector |n|

'''

# FIXME: Move these elsewhere
def split(st,fn=None):
	return map(fn,string.split(st))

def join(list):
	return string.join(map(str,list))

def readFile(filename):
	lines = open(filename,"r").readlines()
	lines = map(string.strip,lines)
	lines = filter(lambda x: len(x)>0,lines)
	lines = filter(lambda x: x[0] != "#",lines)
	return lines

def read(filename):
	lines = readFile(filename)
	n,m,k = split(lines[0],int)
	beta = split(lines[1],float)
	assert len(beta)==n
	V = []
	for i in range(m):
		vRow = split(lines[i+2],float)
		assert len(vRow)==n
		V.append(vRow)
	P = []
	for i in range(m):
		pRow = split(lines[i+2+m],float)
		assert len(pRow)==n
		P.append(pRow)
	return (n,m,k,beta,V,P)	

def save(setting,filename):
	n,m,k,beta,V,P = setting
	f = open(filename,'w')
	f.write("# N,M,K\n")
	f.write("%d %d %d\n"%(n,m,k))
	f.write("# Beta\n")
	f.write(join(beta)+"\n")
	f.write("# V\n")
	for i in range(m):
		f.write(join(V[i])+"\n")
	f.write("# P\n")
	for i in range(m):
		f.write(join(P[i])+"\n")
		
def normalizePerClick(setting):
        ''' Normalize values so that the highest value per click is k '''
   	n,m,k,beta,V,P = setting
        Z = max([ (V[j][i]) for i in range(n) for j in range(m)])
        for i in range(n):
            for j in range(m):
                V[j][i] *= k/Z
	return (n,m,k,beta,V,P)	

def normalizePerImpression(setting):
        ''' Normalize values so that the highest value per impression is k '''
   	n,m,k,beta,V,P = setting
        Z = max([ (V[j][i]*P[j][i]) for i in range(n) for j in range(m)])
        for i in range(n): 
            for j in range(m):
                V[j][i] *= k/Z
	return (n,m,k,beta,V,P)	

def payoffs(setting,assignment=None):
        if assignment == None:
            assignment = optimalAssignment(setting)
	n,m,k,beta,V,P = setting
	po = [(0.0) for i in range(n)]
	for slot in range(len(assignment)):
	    i = assignment[slot]
	    if i != -1:
    		po[i] = (V[slot][i]*P[slot][i])
	return po

import tree_search

class AssignmentNode(tree_search.Node):

    def __init__(self,setting,assignment,random):
        self.X = 1000
        self.setting = setting
        self.assignment = assignment
        self.random = random
    def is_goal(self):
   	n,m,k,beta,V,P = self.setting
        return len(self.assignment)==m
    def children(self):
   	if self.is_goal(): return []
   	n,m,k,beta,V,P = self.setting
   	children = []
   	bidders = range(-1,n)
   	for i in range(3):
            self.random.shuffle(bidders)
        for i in bidders:
            if i == -1 or i not in self.assignment:
                children.append(AssignmentNode(self.setting,self.assignment+[i],self.random))
        return children
    def cost(self):
   	n,m,k,beta,V,P = self.setting
        total = 0.0
        for p in range(len(self.assignment)):
            i = self.assignment[p]
            value = V[p][i]*P[p][i]
            if i == -1: value = 0
            total += self.X -value
        return total
    def heuristic(self):
   	n,m,k,beta,V,P = self.setting
        total = 0.0
        candidates = filter(lambda x:x not in self.assignment,range(n))
        for p in range(len(self.assignment),m):
            value = max([(V[p][i]*P[p][i]) for i in candidates])
            total += min([self.X-0.0,self.X-value])
        return total
    def __str__(self):
        return str(self.assignment)+str(self.cost())
    
def socialWelfare(setting, assignment=None):
        ''' Compute the social welfare of an assignment in the setting
            If no assignment is provided, compute the optimal social welfare '''
        if assignment == None:
            assignment = optimalAssignment(setting)
	n,m,k,beta,V,P = setting
        return sum(payoffs(setting,assignment))
                   
# FIXME: Move these elsewhere
def copyVector(vector):
    return [(element) for element in vector]
def copyMatrix(matrix):
    return [(copyVector(row)) for row in matrix]

def VCG_Payments(setting,seed=None):
    n,m,k,beta,V,P = setting
    effAlloc = optimalAssignment(setting,seed)
    sw = socialWelfare(setting,effAlloc)
    payments = {}
    
    for slot in range(len(effAlloc)):
        i = effAlloc[slot]
        if i == -1: continue
        value = V[slot][i]*P[slot][i]
        swWithI = sw-value
        Vprime = copyMatrix(V)
        for pos in range(len(V)):
            Vprime[pos][i] = 0.0
        swWithoutI = socialWelfare( (n,m,k,beta,Vprime,P) )
        payments[i] = swWithoutI-swWithI
    return payments
    
def VCG_Revenue(setting,seed=None):
    return sum([(payment) for payment in VCG_Payments(setting,seed).values()])

def optimalAssignmentSearch(setting,seed=None):
    ''' Returns the optimal assignment (slot->bidder) '''
    #return __optimalDFS(setting,[])[1]
    import random as r
    r.seed(seed)
    return tree_search.branch_and_bound(AssignmentNode(setting,[],r)).assignment

def optimalAssignmentGreedy(setting):
    ''' Warning: Doesn't work with BenischSadehSandholm!!! '''
    n,m,k,beta,V,P = setting
    assignment = {}
    while len(assignment) < m:
        candidates = filter(lambda x:x not in assignment.values(),range(n))
        slots = filter(lambda x:x not in assignment.keys(),range(m))
        bidderGain = []
        for i in candidates:
            gains = [(0.0,-1)]+[(-V[j][i]*P[j][i],j) for j in slots]
            gains.sort()
            slot = gains[0][1]
            value = gains[0][0]-gains[1][0]
            bidderGain.append( (value,i,slot) )
            #print i,gains
        #print
        bidderGain.sort()
        best,i,slot = bidderGain[0]
        if slot == -1:
            break
        assignment[slot] = i
    print assignment
    for slot in range(m):
        if not assignment.has_key(slot): assignment[slot] = -1
    return [(assignment[slot]) for slot in range(m)]

def __optimalDFS(setting,av):
    ''' Returns the best assignment it can get by extending assignment vector av '''
    n,m,k,beta,V,P = setting    
    if len(av)==m:
        a = dict([(i,av[i]) for i in range(len(av))])
        return socialWelfare(setting,a),a
    scores = []
    for i in range(n):
        if i in av: continue
        childAV = av+[i]
        scores.append(__optimalDFS(setting,childAV))
    return max(scores)

def humanReadible(setting,filename,short=True):
    ''' Writes the setting to a more human-readible HTML format '''
    n,m,k,beta,V,P = setting 
    f = open(filename,'w')
    f.write("<HTML><BODY>\n")
    f.write("<H2>Bidder Quality</H2><TABLE BORDER=1><TR><TD/>\n")
    for i in range(n):
        f.write("<TH>Bidder %d</TH>"%(i))
    f.write("</TR>\n<TR><TH>Beta</TH>")
    for i in range(n):
        f.write("<TD>%.3f</TD>"%(beta[i]))
    f.write("</TR></TABLE>\n<H2>Value per Click</H2><TABLE BORDER=1><TR><TD/>\n")
    for i in range(n):
        f.write("<TH>Bidder %d</TH>"%(i))
    for j in range(m):
        f.write("</TR>\n<TR><TH>Position %d</TH>"%(j))
        for i in range(n):
            f.write("<TD>%.3f</TD>"%(V[j][i]))
    f.write("</TR></TABLE>\n<H2>Click-Through Rate</H2><TABLE BORDER=1><TR><TD/>\n")
    for i in range(n):
        f.write("<TH>Bidder %d</TH>"%(i))
    for j in range(m):
        f.write("</TR>\n<TR><TH>Position %d</TH>"%(j))
        for i in range(n):
            f.write("<TD>%.3f</TD>"%(P[j][i]))
    f.write("</TR></TABLE>\n<H2>Value per Impression</H2><TABLE BORDER=1><TR><TD/>\n")
    for i in range(n):
        f.write("<TH>Bidder %d</TH>"%(i))
    for j in range(m):
        f.write("</TR>\n<TR><TH>Position %d</TH>"%(j))
        for i in range(n):
            f.write("<TD>%.3f</TD>"%(P[j][i]*V[j][i]))
    f.write("</TR></TABLE>\n")
    if short:
        f.close()
        return
    raise "This is no longer working!"
    f.write("<H2>VCG Properties</H2><TABLE BORDER=1><TR><TD/>\n")
    assignment = optimalAssignment(setting)
    for i in range(n):
        f.write("<TH>Bidder %d</TH>"%(i))
    f.write("</TR>\n<TR><TH>Position</TH>")
    for i in range(n):
        rank = ""
        for r in assignment.keys():
            if assignment[r] == i: rank = str(r)
        f.write("<TD>%s</TD>"%(rank))
    f.write("</TR>\n<TR><TH>Revenue</TH>")
    po = payoffs(setting,assignment)
    for i in range(n):
        f.write("<TD>%.3f</TD>"%(po[i]))
    f.write("</TR>\n<TR><TH>Payment</TH>")
    payments = VCG_Payments(setting)
    for i in range(n):
        payment = 0.0
        if payments.has_key(i): payment = payments[i]
        f.write("<TD>%.3f</TD>"%(payment))
    f.write("</TR>\n<TR><TH>Payoff</TH>")
    payments = VCG_Payments(setting)
    print payments
    for i in range(n):
        payment = 0.0
        if payments.has_key(i): payment = payments[i]
        f.write("<TD>%.3f</TD>"%(po[i]-payment))
    f.write("</TR></TABLE>\n")

    f.close()


def EdelmanSchwarzOstrovsky(n,m,k,seed=None):
	''' Makes a position auction setting similar to EdelmanSchwarzOstrovsky's:
	    n = number of bidders, m = number of slots, k = number of bid increments
		Values per click are independent of position
		Beta = advertiser quality
		Click-through factors into beta (advertiser quality) and alpha (slot quality)
	'''
	import random as r
	r.seed(seed)
	b = r.random()
	beta = [b]*n
	alpha = []
	alphaPrime = 1.0
	for i in range(m): 
		a = alphaPrime*r.random()
		alpha.append(a)
		alphaPrime = a
	P = []
	for j in range(m):
		pRow = [(alpha[j]*b) for b in beta]
		P.append(pRow)
	vRow = [(r.random()*k) for i in range(n)]
	V = [(list(vRow)) for j in range(m)]
	return (n,m,k,beta,V,P)

def Varian(n,m,k,seed=None):
	''' Makes a position auction setting similar to Varian's:
	    n = number of bidders, m = number of slots, k = number of bid increments
		Values per click are independent of position
		Beta = advertiser quality
		Click-through factors into beta (advertiser quality) and alpha (slot quality)
	'''
	import random as r
	r.seed(seed)
	beta = [(r.random()) for i in range(n)]
	alpha = []
	alphaPrime = 1.0
	for i in range(m): 
		a = alphaPrime*r.random()
		alpha.append(a)
		alphaPrime = a
	P = []
	for j in range(m):
		pRow = [(alpha[j]*b) for b in beta]
		P.append(pRow)
	vRow = [(r.random()*k) for i in range(n)]
	V = [(list(vRow)) for j in range(m)]
	return (n,m,k,beta,V,P)

from math import sqrt,pi,exp

def __gauss(mu,sigma,x):
    diff = float(x-mu)
    return 1.0/(sigma*sqrt(2*pi)) * exp( - (diff*diff)/(2*sigma*sigma) )

def __rankNormalize(r,m):
    return 1.0-(r)/(m-1.0)

def BenischSadehSandholm(n,m,k,seed=None):
	''' Makes a position auction setting similar to Benisch, Sadeh & Sandholms's:
	    n = number of bidders, m = number of slots, k = number of bid increments
		Values per click are independent of position
		Beta = advertiser quality
		Click-through factors into beta (advertiser quality) and alpha (slot quality)
	'''
	C = 1.0
	import random as r
	r.seed(seed)
	beta = [(1.0) for i in range(n)]
	
	#ALPHA = [0.1,0.0774,0.0666,0.0574,0.0500,0]
	#alpha = ALPHA[:m]
	alpha = [(0.1*math.pow(2.0/3,i/2.0)) for i in range(m)]
	P = []
	for j in range(m):
		pRow = [(alpha[j]*b) for b in beta]
		P.append(pRow)
        mus = []
        for i in range(n):
            if r.random() >= 0.5:
                mus.append(r.random()*0.2+0.8)
            else:
                mus.append(r.random()*0.2+0.4)
	valuePerClick = [(r.random()*120+30) for i in range(n)]
	V = []
	for j in range(m):
            vRow = [(__gauss(mus[i],mus[i]*0.25,__rankNormalize(j,m))*valuePerClick[i]-C) for i in range(n)]
            V.append(vRow)
	return (n,m,k,beta,V,P)

def monotonicVector(n,random):
    vector = [1.0]
    for i in range(1,n):
        vector.append(random.random()*vector[-1])
    return vector

def BlumrosenHartlineNong(n,m,k,seed=None):
	''' Makes a position auction setting similar to Blumrosen, Hartline, & Nong's:
	    n = number of bidders, m = number of slots, k = number of bid increments
		Values per click are independent of position
		Beta = advertiser quality
		Click-through factors into beta (advertiser quality) and alpha (slot quality)
	'''
	import random as r
	r.seed(seed)
	ci = [(r.random()) for i in range(n)]
	cj = monotonicVector(m,r)
	P = []
	for j in range(m):
		pRow = [(cj[j]*ci[i]) for i in range(n)]
		P.append(pRow)
        alphai = [(r.random()) for i in range(n)]
        alphaj = [1.0]
        for i in range(m-1):
            maxAlphaj = alphaj[-1]*cj[i]/cj[i+1]
            nextAlphaj = alphaj[-1]+r.random()*(maxAlphaj-alphaj[-1])*0.3 # FIXME, I need to tune this parameter... 
            alphaj.append(nextAlphaj)
	conversionValues = [(r.random()*k) for i in range(n)]
	V = []
	for j in range(m):
		vRow = [(alphaj[j]*alphai[i]*conversionValues[i]) for i in range(n)]
		V.append(vRow)
	beta = ci
	return (n,m,k,beta,V,P)

distFns = {"ESO":EdelmanSchwarzOstrovsky,"V":Varian,"BSS":BenischSadehSandholm,"BHN":BlumrosenHartlineNong}

if __name__=="__main__":
    from sys import argv
    try:
        filename = argv[1]
        distribution = argv[2]
        n,m,k,seed = map(int,argv[3:])
        assert distribution in distFns
        distFn = distFns[distribution]
        setting = distFn(n,m,k,seed)
        save(setting,filename)
    except:
        print "usage: %s <filename> <distribution> <n> <m> <k> <seed>"%(argv[0])
        print "Distribution is one of",distFns.keys()


def DiscretizeForVCG(setting): # Note: only works for pay-per-click
    round = lambda x:math.floor(x+0.5)
    n,m,k,beta,V,P = setting

    V2 = []
    for j in range(m):
        row2 = []
        for i in range(n):
            
            v = V[j][i]
            v2 = round(v)
            row2.append( v2 )
        V2.append(row2)

    return n,m,k,beta,V2,P

''' Problem: What to do when optimal assignment leaves a slot empty? Answer, leave them empty! '''
''' Problem: My first pass of greedy doesn't work.  Can I design a good DP algorithm? '''


    
def cplexify(setting):
    n,m,k,beta,V,P = setting
    out = "max "
    for i in range(n):
        for j in range(m):
            out += "%fb%dp%d + "%(V[j][i]*P[j][i],i,j)
    out = out[:-3]+"\n\nst\n\n"
    entries = []
    for i in range(n):
        for j in range(m):
            entries.append("b%dp%d"%(i,j))
    for i in range(n):
        entries.append(string.join([("b%dp%d"%(i,j)) for j in range(m)]," + "))
    for j in range(m):
        entries.append(string.join([("b%dp%d"%(i,j)) for i in range(n)]," + "))
    for entry in entries:
        out += "%s >= 0\n\n%s <= 1\n\n"%(entry,entry)
    out += "\n\n\ninteger\n\n"
    for i in range(n):
        for j in range(m):
            out += "b%dp%d "%(i,j)

    out = string.replace(out,"+ -","-")
    return out+"\n"

def tempfileName():
    import tempfile
    f = tempfile.NamedTemporaryFile()
    name = f.name
    f.close()
    return name

def optimalAssignmentCPLEX(setting,seed=None):
    tmpFilename = tempfileName()
    n,m,k,beta,V,P = setting
    open(tmpFilename+".lp",'w').write(cplexify(setting))
    print cplexify(setting)
    import os,re
    open(tmpFilename+".i",'w').write("read "+tmpFilename+".lp\noptimize\ndisplay solution variables *\nquit\n")
    #assert os.system("cplex <"+tmpFilename+".i 1>"+tmpFilename+".o 2>/dev/null")==0
    assert os.system("cplex <"+tmpFilename+".i 1>"+tmpFilename+".o")==0
    results = open(tmpFilename+".o",'r').read()
    print results
    variables = re.findall("b[0-9]+p[0-9]+",results)
    assignment = [-1]*m
    for var in variables: # Assume these are 1.0
        i,j = map(int,re.split("[bp]",var)[1:])
        assignment[j] = i
    try:os.remove(tmpFilename)
    except:pass
    os.remove(tmpFilename+'.lp')
    os.remove(tmpFilename+'.i')
    os.remove(tmpFilename+'.o')
    return assignment

optimalAssignment = optimalAssignmentCPLEX

if False:
    s = BlumrosenHartlineNong(5,3,5,47)
    normalizePerClick(s)
    humanReadible(s,"Tmp1.html")
    s2 = DiscretizeForVCG(s)
    humanReadible(s2,"Tmp2.html")
    
    print socialWelfare(s)
    print "revenue",VCG_Revenue(s)
    for i in range(1,0):
        s2 = DiscretizeForVCG(s)
        #oa2 = optimalAssignmentSearch(s2)
        #print "search",socialWelfare(s,oa2),oa2
        oa2 = optimalAssignmentCPLEX(s2,'tmpfile')
        print "cplex",socialWelfare(s,oa2),oa2
        print "revenue",VCG_Revenue(s2),VCG_Payments(s2)
